package com.loafle.overflow.module.member.service;

import com.loafle.overflow.commons.model.SessionMetadata;
import com.loafle.overflow.commons.stereotype.WebappAPI;
import com.loafle.overflow.commons.utils.EmailSender;
import com.loafle.overflow.module.apikey.model.ApiKey;
import com.loafle.overflow.module.apikey.service.ApiKeyService;
import com.loafle.overflow.module.domain.model.Domain;
import com.loafle.overflow.module.domain.model.DomainMember;
import com.loafle.overflow.module.domain.service.DomainMemberService;
import com.loafle.overflow.module.email.service.EmailAuthService;
import com.loafle.overflow.module.member.dao.MemberDAO;
import com.loafle.overflow.module.member.exception.*;
import com.loafle.overflow.module.member.model.Member;
import com.loafle.overflow.module.meta.model.MetaMemberStatus;
import com.loafle.overflow.module.probe.model.Probe;
import com.loafle.overflow.module.probe.service.ProbeService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.mail.MailException;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.stereotype.Service;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Created by geek on 17. 6. 28.
 */
@Service("MemberService")
public class MemberService {

    @Autowired
    private MemberDAO memberDAO;

    @Autowired
    private EmailAuthService emailAuthService;

    @Autowired
    private EmailSender emailSender;

    @Autowired
    private ApiKeyService apiKeyService;

    @Autowired
    private DomainMemberService domainMemberService;

    @Autowired
    private ProbeService probeService;

    private BCryptPasswordEncoder passwordEncoder = new BCryptPasswordEncoder();

    public DomainMember signin(String signinId, String signinPw) {
        Member m = this.memberDAO.findByEmail(signinId);

        if ( null == m ) {
            throw new SignInIdNotExistException();
        }

        if ( m.getStatus().getId() == 1 ) {
            throw new EmailNotConfirmedException();
        }

        Boolean match = passwordEncoder.matches(signinPw, m.getPw());
        if(!match) {
            if (m.getSigninFailCount() > 10) {
                throw new SigninOverFailedException("Login failed 10 times");
            }
            m.setSigninFailCount(m.getSigninFailCount()+1);
            this.modify(m);
            throw new SignInPwNotMatchException();
        }

        m.setSigninFailCount(0);
        this.modify(m);

        DomainMember dm = domainMemberService.readByMemberEmail(m.getEmail());

        // Todo Signin History
        return dm;
    }

    public Member signup(Member member, String pw) throws UnsupportedEncodingException {

        Member isMember = this.memberDAO.findByEmail(member.getEmail());

        if (null != isMember && isMember.getId() > 0) {
            throw new JoinedEmailException();
        }

        boolean checkPass = this.isPasswordStrong(pw);

        if (!checkPass) {
            throw new PasswordNotStrongException(
                    "Passwords must contain at least one uppercase letter, " +
                            "special character, lowercase letter, and number, " +
                            "and must be at least 6 characters long.");
        }
        member.setPw(passwordEncoder.encode(pw));

        if (member.getStatus() == null) {
            member.setStatus(new MetaMemberStatus((short) 1));
        }

        Member resMember = this.memberDAO.save(member);

        try {
            this.emailAuthService.sendEmailByMember(member.getId(), member.getEmail());
        } catch (MailException e) {
            // Todo ReSend Mail
            e.printStackTrace();
        }

        return resMember;
    }

    public Member sendEmailForPassword(String email) throws UnsupportedEncodingException {
        Member member = this.memberDAO.findByEmail(email);

        if (null == member) {
            throw new SignInIdNotExistException();
        }

        try {
            this.emailAuthService.sendEmailResetPassword(member);
        } catch (MailException e) {
            // Todo ReSend Mail
            e.printStackTrace();
        }

        return member;
    }

    public Member resetPassword(String encodeEmail, String pw) throws UnsupportedEncodingException {
        String deStr = URLDecoder.decode(encodeEmail, "UTF-8");

        String deEmail = this.emailSender.decrypt(deStr);
        Member member = this.memberDAO.findByEmail(deEmail);

        if (null == member) {
            throw new SignInIdNotExistException();
        }

        boolean checkPass = this.isPasswordStrong(pw);

        if (!checkPass) {
            throw new PasswordNotStrongException(
                    "Passwords must contain at least one uppercase letter, " +
                            "special character, lowercase letter, and number, " +
                            "and must be at least 6 characters long.");
        }
        member.setPw(passwordEncoder.encode(pw));

        return this.modify(member);
    }

    public void signout(Member member) {
        // Todo websocket session remove
    }

    @WebappAPI
    public Member modify(Member member, String pw) {
        String email = SessionMetadata.getTargetID();
        Member preMember = this.memberDAO.findByEmail(member.getEmail());

        if (null != pw && !pw.equals("")) {
            boolean checkPass = this.isPasswordStrong(pw);

            if (!checkPass) {
                throw new PasswordNotStrongException(
                        "Passwords must contain at least one uppercase letter, " +
                                "special character, lowercase letter, and number, " +
                                "and must be at least 6 characters long.");
            }

            Boolean match = passwordEncoder.matches(member.getPw(), preMember.getPw());
            if(!match) {
                member.setPw(passwordEncoder.encode(pw));
            }
        } else {
            member.setPw(preMember.getPw());
        }

        if (member.getStatus() == null || member.getStatus().getId() < 0) {
            member.getStatus().setId(preMember.getStatus().getId());
        }
        return this.modify(member);
    }

    private Member modify(Member member) {
        return this.memberDAO.save(member);
    }

    public Member confirmPw(String signinId, String signinPw) {
        Member preMember = this.memberDAO.findByEmail(signinId);

        String encodePw = passwordEncoder.encode(signinPw);
        Boolean match = passwordEncoder.matches(encodePw, preMember.getPw());

        if (!match) {
            throw new SignInPwNotMatchException();
        }

        return preMember;
    }

    public Member forgotPassword(String signinId, String newPw) {
        Member preMember = this.memberDAO.findByEmail(signinId);

        if (null == preMember) {
            throw new SignInIdNotExistException();
        }

        Member cMember = this.modify(preMember, newPw);

        return cMember;
    }

    public Member read(long memberId) {
        if (memberId <= 0) {
            throw new SignInIdNotExistException();
        }

        Member resMember = this.memberDAO.findOne(memberId);
        return resMember;
    }

    @WebappAPI
    public void withdrawal(Member member) {
        String email = SessionMetadata.getTargetID();

        // Todo DB delete?
    }

    public void emailConfirm(Member member) {

    }

    public List<Member> readAllByProbeKey(String probeKey) {

        Probe probe = this.probeService.readByProbeKey(probeKey);

        if(probe == null) {
            return null;
        }

        return this.domainMemberService.readAllMemberByDomain(probe.getDomain());
    }

    public List<Member> readAllByApiKey(String apikey) {

        ApiKey apiKey = this.apiKeyService.readByApiKey(apikey);

        if(apiKey == null) {
            return null;
        }

        return this.domainMemberService.readAllMemberByDomain(apiKey.getDomain());
    }

    public List<Member> readAllByDomain(Domain domain) {

        return this.domainMemberService.readAllMemberByDomain(domain);
    }

    public List<Member> readAllByDomainID(final long domainID) {

        return this.domainMemberService.readAllMemberByDomainID(domainID);
    }

    private static final String PASSWORD_REGEXP =  "(" +
            "(?=.*[a-z])" +
            "(?=.*\\d)" +
            "(?=.*[A-Z])" +
            "(?=.*[!@#$%^&*()_+\\-=\\[\\]{};':\"\\\\|,.<>\\/?])" +
            "." +
            "{6,40}" +
            ")";
    private Pattern pattern = Pattern.compile(PASSWORD_REGEXP);

    protected boolean isPasswordStrong(String pass) {
        Matcher m = pattern.matcher(pass);
        return m.matches();
    }
}
